
#include <ntddk.h>
#include "pciscc_drv.h"
#include "pciscc_ioctl.h"
#include "pciscc_flex.h"

#define	CHIP_TIMEOUT	1000000		/* chip command timeout */
#define	XTAL_FREQ		19660000	/* on board crystal freq. */
#define	MIN_FLAGS		32			/* minimum delay/tail bits */

NTSTATUS DriverEntry(
	IN PDRIVER_OBJECT  DriverObject,
	IN PUNICODE_STRING RegistryPath
	)
{
	UNICODE_STRING NtDeviceName;
	UNICODE_STRING Win32DeviceName;

	// Initialize the driver object dispatch table.
	// NT sends requests to these routines.
	DriverObject->MajorFunction[IRP_MJ_CREATE]			= pciscc_create;
	DriverObject->MajorFunction[IRP_MJ_CLOSE]			= pciscc_close;
	DriverObject->MajorFunction[IRP_MJ_READ]			= pciscc_read;
	DriverObject->MajorFunction[IRP_MJ_WRITE]			= pciscc_write;
	DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]	= pciscc_ioctl;
	DriverObject->MajorFunction[IRP_MJ_PNP]				= pciscc_pnp;
    DriverObject->MajorFunction[IRP_MJ_POWER]			= pciscc_power;
    DriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL]	= pciscc_sysctl;
	DriverObject->DriverUnload							= pciscc_unload;
    DriverObject->DriverExtension->AddDevice			= pciscc_add_device;

    return STATUS_SUCCESS;
}

NTSTATUS pciscc_add_device(
    IN PDRIVER_OBJECT DriverObject,
    IN PDEVICE_OBJECT PhysicalDeviceObject
    )
{
	NTSTATUS Status;
	PDEVICE_OBJECT DeviceObject;
	PLOCAL_DEVICE_INFO pLDI;
	UNICODE_STRING NtDeviceName;
	UNICODE_STRING Win32DeviceName;
	
	/* Get UNICODE name for device.*/
	RtlInitUnicodeString(&NtDeviceName, PCISCC_NT_NAME_0);
	RtlInitUnicodeString(&Win32DeviceName, PCISCC_W32_NAME_0);

	/* Create our device. */
	Status = IoCreateDevice(
					DriverObject,
					sizeof(LOCAL_DEVICE_INFO),
					&NtDeviceName,
					FILE_DEVICE_UNKNOWN,
					0,
					FALSE,
					&DeviceObject
					);

	/* Card 0 already managed by any driver ? */
	if (!NT_SUCCESS(Status)) {

		/* Get UNICODE name for device.*/
		RtlInitUnicodeString(&NtDeviceName, PCISCC_NT_NAME_1);
		RtlInitUnicodeString(&Win32DeviceName, PCISCC_W32_NAME_1);

		/* Create our device. */
		Status = IoCreateDevice(
						DriverObject,
						sizeof(LOCAL_DEVICE_INFO),
						&NtDeviceName,
						FILE_DEVICE_UNKNOWN,
						0,
						FALSE,
						&DeviceObject
						);

		if (!NT_SUCCESS(Status))
			return Status;
	}

	/* Initialize the local driver info */
	pLDI = (PLOCAL_DEVICE_INFO)DeviceObject->DeviceExtension;

	/* Clear local device info memory */
	RtlZeroMemory(pLDI, sizeof(LOCAL_DEVICE_INFO));

	DeviceObject->Flags |= DO_BUFFERED_IO;
	DeviceObject->AlignmentRequirement = FILE_BYTE_ALIGNMENT;
	
	InitializeListHead(&pLDI->RxQueue);
	InitializeListHead(&pLDI->RxPending);
	KeInitializeSpinLock(&pLDI->Lock);
	IoInitializeRemoveLock(&pLDI->RemoveLock,PCISCC_TAG,1,5);

	KeInitializeDpc(&pLDI->TxResetDpc, pciscc_dpc_txreset, DeviceObject);
	KeInitializeDpc(&pLDI->IsrDpc, pciscc_dpc_isr, DeviceObject);

	/* Create symbolic device link */
	Status = IoCreateSymbolicLink( &Win32DeviceName, &NtDeviceName );
	if (!NT_SUCCESS(Status))
	{
		IoDeleteDevice(DeviceObject);
		return Status;
	}

	/* Interface to lower driver */
	pLDI->NextLowerDrv = IoAttachDeviceToDeviceStack(	DeviceObject,
														PhysicalDeviceObject);
	if (pLDI->NextLowerDrv == NULL) {
		IoDeleteSymbolicLink(&Win32DeviceName);
		IoDeleteDevice(DeviceObject);
		Status = STATUS_NO_SUCH_DEVICE;
	}

	/* Copy device name to device info structure */
	RtlCopyUnicodeString(&pLDI->Win32DeviceName, &Win32DeviceName);

	/* Device init complete */
	DeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;
	return Status;
}

NTSTATUS pciscc_pnp(
    IN PDEVICE_OBJECT pDO,
    IN PIRP pIrp
    )
{
    PIO_STACK_LOCATION	pIrpStack;
    KEVENT				Event;
    PLOCAL_DEVICE_INFO	pLDI;
	NTSTATUS			Status = STATUS_SUCCESS;

	/* get control parameter */
	pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;
	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

	/* acquire remove lock*/
	Status = IoAcquireRemoveLock (&pLDI->RemoveLock, pIrp);
	if (!NT_SUCCESS (Status)) {
		pIrp->IoStatus.Status = Status;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
		return Status;
	}

	switch (pIrpStack->MinorFunction) {
	case IRP_MN_START_DEVICE:
		/* The device is starting */
		IoCopyCurrentIrpStackLocationToNext(pIrp);
		KeInitializeEvent(&Event,NotificationEvent,FALSE);

		IoSetCompletionRoutine(pIrp,(PIO_COMPLETION_ROUTINE) pciscc_complete, 
                               &Event,TRUE,TRUE,TRUE);

		Status = IoCallDriver(pLDI->NextLowerDrv, pIrp);
		if (Status == STATUS_PENDING)
			KeWaitForSingleObject(&Event,Executive,KernelMode,FALSE,NULL);
		if (NT_SUCCESS(Status) && NT_SUCCESS(pIrp->IoStatus.Status))
		{
			Status = pciscc_2k_config(pDO, pIrp);
			if (NT_SUCCESS (Status))
				Status = pciscc_chip_open(pLDI);
        }
		pIrp->IoStatus.Status = Status;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
        break;

    case IRP_MN_QUERY_STOP_DEVICE:
    case IRP_MN_STOP_DEVICE:
		pIrp->IoStatus.Status = STATUS_UNSUCCESSFUL;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
		break;

    case IRP_MN_QUERY_REMOVE_DEVICE:
    case IRP_MN_CANCEL_REMOVE_DEVICE: 
	case IRP_MN_CANCEL_STOP_DEVICE:
		pIrp->IoStatus.Status = STATUS_SUCCESS;
		IoSkipCurrentIrpStackLocation(pIrp);
		Status = IoCallDriver(pLDI->NextLowerDrv, pIrp);
		break;

    case IRP_MN_SURPRISE_REMOVAL:
		/* Stop chip activity */
		pciscc_chip_close(pLDI);

		pIrp->IoStatus.Status = STATUS_SUCCESS;
		IoSkipCurrentIrpStackLocation(pIrp);
		Status = IoCallDriver(pLDI->NextLowerDrv, pIrp);
		break;
        
    case IRP_MN_REMOVE_DEVICE:
		/* Stop chip activity */
		pciscc_chip_close(pLDI);

		IoDeleteSymbolicLink(&pLDI->Win32DeviceName);
		IoReleaseRemoveLockAndWait(&pLDI->RemoveLock, pIrp);

		pIrp->IoStatus.Status = STATUS_SUCCESS;
		IoSkipCurrentIrpStackLocation(pIrp);
		Status = IoCallDriver(pLDI->NextLowerDrv, pIrp);

		IoDetachDevice(pLDI->NextLowerDrv);
		IoDeleteDevice(pDO);
		return Status;
                
    default:
        //
        // Please see PnP documentation for use of these IRPs.
        //
        IoSkipCurrentIrpStackLocation (pIrp);
        Status = IoCallDriver(pLDI->NextLowerDrv, pIrp);
        break;
    }

    IoReleaseRemoveLock(&pLDI->RemoveLock, pIrp);       
    return Status;
}

NTSTATUS pciscc_power(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp
    )
{
	PLOCAL_DEVICE_INFO   pLDI;
    
	pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;

	PoStartNextPowerIrp(pIrp);
	IoSkipCurrentIrpStackLocation(pIrp);
	return PoCallDriver(pLDI->NextLowerDrv, pIrp);
}

NTSTATUS pciscc_sysctl(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp
	)
{
	PLOCAL_DEVICE_INFO pLDI;

	pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;
	IoSkipCurrentIrpStackLocation(pIrp);
	return IoCallDriver(pLDI->NextLowerDrv,pIrp);
}

NTSTATUS pciscc_create(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp             
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;

	pLDI->OpenCnt++;
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

NTSTATUS pciscc_close(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp             
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	KIRQL OldIrql;
	ULONG chan;

	pLDI->OpenCnt--;
	if (!pLDI->OpenCnt) {
		KeAcquireSpinLock(&pLDI->Lock,&OldIrql);
		for (chan = 0; chan < 4; chan++)
			pciscc_channel_close(pLDI,chan);
		KeReleaseSpinLock(&pLDI->Lock,OldIrql);
	}
	pIrp->IoStatus.Status = STATUS_SUCCESS;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

NTSTATUS pciscc_read(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	PIO_STACK_LOCATION pIrpStack;
	PLIST_ENTRY pEntry;
	L1FRAME *rx_frame;
	KIRQL OldIrql;
	ULONG Count;
	ULONG Result = 0;

	/* Get read buffer length */
	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	Count = pIrpStack->Parameters.Read.Length;
	if (Count > sizeof(L1FRAME))
		Count = sizeof(L1FRAME);

	KeAcquireSpinLock(&pLDI->Lock,&OldIrql);
	/* RX frame list empty ? */
	if (!IsListEmpty(&pLDI->RxQueue)) {
		pEntry = RemoveHeadList(&pLDI->RxQueue);
		rx_frame = CONTAINING_RECORD(pEntry, L1FRAME, list_entry);
		RtlCopyMemory(pIrp->AssociatedIrp.SystemBuffer, rx_frame, Count);
		pciscc_free(rx_frame);
		Result = Count;
	}
	if ((pLDI->ReadMode == 1) && (Result == 0)) {
		pIrp->IoStatus.Status = STATUS_PENDING;
		IoMarkIrpPending(pIrp);
		IoSetCancelRoutine(pIrp, pciscc_cancel);
		InsertTailList(&pLDI->RxPending, &pIrp->Tail.Overlay.ListEntry);
		KeReleaseSpinLock(&pLDI->Lock,OldIrql);

		return STATUS_PENDING;
	}
	KeReleaseSpinLock(&pLDI->Lock,OldIrql);

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}

NTSTATUS pciscc_write(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	PIO_STACK_LOCATION pIrpStack;
	XMIT_SYNC_INFO TxData;
	KIRQL OldIrql;
	ULONG Count;
	ULONG Result = 0;

	/* Get write buffer length */
	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	Count = pIrpStack->Parameters.Write.Length;
	if (Count > sizeof(L1FRAME))
		Count = sizeof(L1FRAME);

	KeAcquireSpinLock(&pLDI->Lock,&OldIrql);
	/* TX frame allocated ? */
	if (TxData.tx_frame = pciscc_alloc(sizeof(L1FRAME))) {
		RtlCopyMemory(TxData.tx_frame, pIrp->AssociatedIrp.SystemBuffer, Count);
		TxData.pLDI = pLDI;
		if (KeSynchronizeExecution(pLDI->IrqObject,pciscc_xmit_sync,&TxData))
			Result = Count;
		else
			pciscc_free(TxData.tx_frame);
	}
	KeReleaseSpinLock(&pLDI->Lock,OldIrql);

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = Result;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	
	return STATUS_SUCCESS;
}

NTSTATUS pciscc_ioctl(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	PIO_STACK_LOCATION pIrpStack;
	CALIB_SYNC_INFO CalibInfo;
	NTSTATUS Status;
	KIRQL OldIrql;
	PULONG pIOBuffer;
	ULONG RdCnt, WrCnt;
	ULONG Result = 0;

	/* Get buffer length */
	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
	pIOBuffer = (PULONG) pIrp->AssociatedIrp.SystemBuffer;

	/* Get output buffer size */
	RdCnt = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	WrCnt = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;

	/* Set default return status */
	Status = STATUS_INVALID_BUFFER_SIZE;

	KeAcquireSpinLock(&pLDI->Lock,&OldIrql);
	switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode)
	{
		case PCISCC_SET_PARAM_0:
			if (WrCnt < 2 * sizeof(ULONG))
				break;
			Status = pciscc_channel_open(pLDI,0,pIOBuffer[0],pIOBuffer[1]);
			break;

		case PCISCC_SET_PARAM_1:
			if (WrCnt < 2 * sizeof(ULONG))
				break;
			Status = pciscc_channel_open(pLDI,1,pIOBuffer[0],pIOBuffer[1]);
			break;

		case PCISCC_SET_PARAM_2:
			if (WrCnt < 2 * sizeof(ULONG))
				break;
			Status = pciscc_channel_open(pLDI,2,pIOBuffer[0],pIOBuffer[1]);
			break;

		case PCISCC_SET_PARAM_3:
			if (WrCnt < 2 * sizeof(ULONG))
				break;
			Status = pciscc_channel_open(pLDI,3,pIOBuffer[0],pIOBuffer[1]);
			break;

		case PCISCC_SET_CALIB_0:
			if (WrCnt < sizeof(ULONG))
				break;
			CalibInfo.pLDI = pLDI;
			CalibInfo.chan = 0;
			CalibInfo.time = pIOBuffer[0];
			KeSynchronizeExecution(pLDI->IrqObject,pciscc_calib_sync,&CalibInfo);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_SET_CALIB_1:
			if (WrCnt < sizeof(ULONG))
				break;
			CalibInfo.pLDI = pLDI;
			CalibInfo.chan = 1;
			CalibInfo.time = pIOBuffer[0];
			KeSynchronizeExecution(pLDI->IrqObject,pciscc_calib_sync,&CalibInfo);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_SET_CALIB_2:
			if (WrCnt < sizeof(ULONG))
				break;
			CalibInfo.pLDI = pLDI;
			CalibInfo.chan = 2;
			CalibInfo.time = pIOBuffer[0];
			KeSynchronizeExecution(pLDI->IrqObject,pciscc_calib_sync,&CalibInfo);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_SET_CALIB_3:
			if (WrCnt < sizeof(ULONG))
				break;
			CalibInfo.pLDI = pLDI;
			CalibInfo.chan = 3;
			CalibInfo.time = pIOBuffer[0];
			KeSynchronizeExecution(pLDI->IrqObject,pciscc_calib_sync,&CalibInfo);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_STATE_0:
			if (RdCnt < sizeof(ULONG))
				break;
			pIOBuffer[0] = pciscc_state(pLDI,0);
			Result = sizeof(ULONG);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_STATE_1:
			if (RdCnt < sizeof(ULONG))
				break;
			pIOBuffer[0] = pciscc_state(pLDI,1);
			Result = sizeof(ULONG);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_STATE_2:
			if (RdCnt < sizeof(ULONG))
				break;
			pIOBuffer[0] = pciscc_state(pLDI,2);
			Result = sizeof(ULONG);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_STATE_3:
			if (RdCnt < sizeof(ULONG))
				break;
			pIOBuffer[0] = pciscc_state(pLDI,3);
			Result = sizeof(ULONG);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_INFO_0:
			if (RdCnt > sizeof(L1_STATISTICS))
				RdCnt = sizeof(L1_STATISTICS);
			RtlCopyMemory(pIOBuffer, &pLDI->Channel[0].stats, RdCnt);
			Result = RdCnt;
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_INFO_1:
			if (RdCnt < sizeof(L1_STATISTICS))
				RdCnt = sizeof(L1_STATISTICS);
			RtlCopyMemory(pIOBuffer, &pLDI->Channel[1].stats, RdCnt);
			Result = RdCnt;
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_INFO_2:
			if (RdCnt < sizeof(L1_STATISTICS))
				RdCnt = sizeof(L1_STATISTICS);
			RtlCopyMemory(pIOBuffer, &pLDI->Channel[2].stats, RdCnt);
			Result = RdCnt;
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_INFO_3:
			if (RdCnt < sizeof(L1_STATISTICS))
				RdCnt = sizeof(L1_STATISTICS);
			RtlCopyMemory(pIOBuffer, &pLDI->Channel[3].stats, RdCnt);
			Result = RdCnt;
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_NONBLOCK_READ:
			pLDI->ReadMode = 0;
			pciscc_clear_rxpending(pLDI);
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_BLOCK_READ:
			pLDI->ReadMode = 1;
			Status = STATUS_SUCCESS;
			break;

		case PCISCC_GET_OPEN:
			if (RdCnt < sizeof(ULONG))
				break;
			pIOBuffer[0] = pLDI->OpenCnt;
			Result = sizeof(ULONG);
			Status = STATUS_SUCCESS;
			break;

		default:
			Status = STATUS_NOT_IMPLEMENTED;
			break;
	}
	KeReleaseSpinLock(&pLDI->Lock,OldIrql);

	pIrp->IoStatus.Status = Status;
	pIrp->IoStatus.Information = Result;
	IoCompleteRequest(pIrp, IO_NO_INCREMENT);

	return Status;
}

/* unload driver */
VOID pciscc_unload(
	IN PDRIVER_OBJECT DriverObject
	)
{
	DbgPrint("PCISCC: Driver unloaded\n");
}

/* completion routine */
NTSTATUS  pciscc_complete(
    IN PDEVICE_OBJECT pDO,
    IN PIRP pIrp,
    IN PVOID Context
    )
{
    PKEVENT	pEvent;
    pEvent = (PKEVENT) Context;

    if (pIrp->PendingReturned)
		IoMarkIrpPending(pIrp);
	KeSetEvent(pEvent, 0, FALSE);
	return STATUS_MORE_PROCESSING_REQUIRED;
}

/* pending IRP cancel routine */
VOID pciscc_cancel(
	IN PDEVICE_OBJECT pDO,
	IN PIRP pIrp
	)
{
	PLOCAL_DEVICE_INFO pLDI = (PLOCAL_DEVICE_INFO)pDO->DeviceExtension;
	KIRQL OldIrql;

	IoReleaseCancelSpinLock(pIrp->CancelIrql);

	KeAcquireSpinLock(&pLDI->Lock,&OldIrql);
    if (!IsListEmpty(&pIrp->Tail.Overlay.ListEntry))
	{
		RemoveEntryList(&pIrp->Tail.Overlay.ListEntry);
		pIrp->IoStatus.Information = 0;
		pIrp->IoStatus.Status = STATUS_CANCELLED;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	}
	KeReleaseSpinLock(&pLDI->Lock,OldIrql);
}

/* ---------------------------------------------------------------- */

/* clear RX frame list on driver shutdown */
VOID pciscc_clear_rxlist(
	IN PLOCAL_DEVICE_INFO pLDI
	)
{
	PLIST_ENTRY pEntry;
	L1FRAME *l1_frame;

	while (!IsListEmpty(&pLDI->RxQueue))
	{
		pEntry = RemoveHeadList(&pLDI->RxQueue);
		l1_frame = CONTAINING_RECORD(pEntry,L1FRAME,list_entry);
		pciscc_free(l1_frame);
	}
}

VOID pciscc_clear_rxpending(
	IN PLOCAL_DEVICE_INFO pLDI
	)
{
	PLIST_ENTRY pEntry;
	PIRP pIrp;

    while (!IsListEmpty(&pLDI->RxPending))
	{
		pEntry = RemoveHeadList(&pLDI->RxPending);
		pIrp = CONTAINING_RECORD(pEntry,IRP,Tail.Overlay.ListEntry);
		pIrp->IoStatus.Information = 0;
		pIrp->IoStatus.Status = STATUS_CANCELLED;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	}
}

/* ---------------------------------------------------------------- */

/* allocate memory at phys. address <= 0xffffffff */
PVOID pciscc_alloc(
	IN ULONG uSize
	)
{
	PHYSICAL_ADDRESS MaxAddr;

	MaxAddr.LowPart = 0xffffffff;
	MaxAddr.HighPart = 0;

	return MmAllocateContiguousMemory(uSize,MaxAddr);
}

/* release previously allocated memory */
VOID pciscc_free(
	IN PVOID BaseAddr
	)
{
	MmFreeContiguousMemory(BaseAddr);
}

/* 32 bit memory read operation */
ULONG ReadL(
	IN PLOCAL_DEVICE_INFO pLDI, 
	IN ULONG Offset
	)
{
	ULONG BaseAddr = (ULONG)pLDI->MappedAddress;
	return READ_REGISTER_ULONG((PVOID)(BaseAddr + Offset));
}

/* 32 bit memory write operation */
VOID WriteL(
	IN PLOCAL_DEVICE_INFO pLDI, 
	IN ULONG Offset,
	IN ULONG Value
	)
{
	ULONG BaseAddr = (ULONG)pLDI->MappedAddress;
	WRITE_REGISTER_ULONG((PVOID)(BaseAddr + Offset), Value);
}

/* Calculate phys. bus address */
ULONG VirtToPhys(
	IN PVOID Address
	)
{
	PHYSICAL_ADDRESS PhysAddr = MmGetPhysicalAddress(Address);
	return PhysAddr.LowPart;
}

/* ---------------------------------------------------------------- */

/* Win2000 PNP config */
NTSTATUS pciscc_2k_config(
    IN PDEVICE_OBJECT pDO,
    IN PIRP pIrp
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	PIO_STACK_LOCATION pIrpStack;
    PCM_RESOURCE_LIST ResourceList;
    PCM_RESOURCE_LIST ResourceListTranslated;
    PCM_PARTIAL_RESOURCE_LIST PartialResourceList;
	ULONG i;

	BOOLEAN	AddressAssigned = FALSE;

	pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

	ResourceList = pIrpStack->Parameters.StartDevice.AllocatedResources;
	ResourceListTranslated = pIrpStack->Parameters.StartDevice.AllocatedResourcesTranslated;

	if ((ResourceList == NULL) || (ResourceListTranslated == NULL))
		return STATUS_INSUFFICIENT_RESOURCES;

	PartialResourceList = &ResourceListTranslated->List[0].PartialResourceList;

	for (i = 0; i < PartialResourceList->Count; i++)
	{
		if (PartialResourceList->PartialDescriptors[i].Type == CmResourceTypeMemory && !AddressAssigned) 
		{
			pLDI->Address = PartialResourceList->PartialDescriptors[i].u.Memory.Start;
			pLDI->AddrLen = PartialResourceList->PartialDescriptors[i].u.Memory.Length;
			AddressAssigned = TRUE;
		}

		if (PartialResourceList->PartialDescriptors[i].Type == CmResourceTypeInterrupt)
		{
			pLDI->IrqLevel = (KIRQL)PartialResourceList->PartialDescriptors[i].u.Interrupt.Level;
			pLDI->IrqVector = PartialResourceList->PartialDescriptors[i].u.Interrupt.Vector;
			pLDI->IrqAffinity = PartialResourceList->PartialDescriptors[i].u.Interrupt.Affinity;
		}
	}

	DbgPrint("PCISCC: PnP reports PEB20534 (DSCC4) chip at address 0x%X\n",pLDI->Address.LowPart);
	return STATUS_SUCCESS;
}

NTSTATUS pciscc_init_queues(
	IN PLOCAL_DEVICE_INFO pLDI,
	IN ULONG chan
	)
{
	struct rx_desc *first_rdp = NULL;
	struct rx_desc *curr_rdp = NULL;
	struct rx_desc *last_rdp = NULL;
	struct tx_desc *first_tdp = NULL;
	struct tx_desc *curr_tdp = NULL;
	struct tx_desc *last_tdp = NULL;
	PCHANNEL_INFO pChan = &pLDI->Channel[chan];
	ULONG err = 0;
	ULONG i;

	/* allocate TX/RX interrupt queues */
	pChan->iq_rx = pLDI->DmaBuffers->iq_rx[chan];
	pChan->iq_tx = pLDI->DmaBuffers->iq_tx[chan];
	RtlZeroMemory(pChan->iq_rx, 4*CFG_IQLEN);
	RtlZeroMemory(pChan->iq_tx, 4*CFG_IQLEN);
	pChan->iq_rx_next = pChan->iq_rx;
	pChan->iq_tx_next = pChan->iq_tx;

	/*
	 * Prepare circular RX and TX descriptor queues ("FIFO" rings)
	 * Attention:
	 * This beast gets _very_ angry if you try to hand it a
	 * descriptor with a data length of 0. In fact it crashes
	 * the system by asserting /SERR or something.
	 */

	/* allocate queues */
	pChan->dq_rx = pLDI->DmaBuffers->dq_rx[chan];
	pChan->dq_tx = pLDI->DmaBuffers->dq_tx[chan];
	RtlZeroMemory(pChan->dq_rx,sizeof(struct rx_desc) * CFG_RX_DESC);
	RtlZeroMemory(pChan->dq_tx,sizeof(struct tx_desc) * CFG_TX_DESC);

	/* RX queue */
	first_rdp = pChan->dq_rx;
	for (i=0, curr_rdp = first_rdp; i < CFG_RX_DESC; i++, curr_rdp++) {
		if (i > 0) {
			curr_rdp->prev = last_rdp;
			last_rdp->next = curr_rdp;
			last_rdp->nextptr = VirtToPhys(curr_rdp);
		}
		curr_rdp->l1_frame = pciscc_alloc(sizeof(L1FRAME));
		if (curr_rdp->l1_frame == NULL) err++;
		curr_rdp->dataptr = VirtToPhys(curr_rdp->l1_frame->frame);
		curr_rdp->flags = CFG_MTU*NO;
		curr_rdp->feptr = 0;
		curr_rdp->result = 0;
		last_rdp = curr_rdp;
	}
	if (err) {
		for (i=0; i < CFG_RX_DESC; i++)
		{
			if (pChan->dq_rx[i].l1_frame)
				pciscc_free(pChan->dq_rx[i].l1_frame);
		}
		return STATUS_NO_MEMORY;
	}
	last_rdp->next = first_rdp;			/* close circular structure */
	last_rdp->nextptr = VirtToPhys(first_rdp);
	first_rdp->prev = last_rdp;
	pChan->dq_rx_next = first_rdp;		/* first descriptor to be processed = "first" descriptor in chain */

	/* TX queue */
	first_tdp = pChan->dq_tx;
	for (i=0, curr_tdp = first_tdp; i < CFG_TX_DESC; i++, curr_tdp++) {
		if (i > 0) {
			curr_tdp->prev = last_tdp;
			last_tdp->next = curr_tdp;
			last_tdp->nextptr = VirtToPhys(curr_tdp);
		}
		curr_tdp->l1_frame = NULL;
		curr_tdp->dataptr = VirtToPhys(pLDI->DmaBuffers->dummy);
		curr_tdp->flags = FE | (8*NO);
		curr_tdp->result = 0;
		last_tdp = curr_tdp;
	}
	last_tdp->next = first_tdp;			/* close circular structure */
	last_tdp->nextptr = VirtToPhys(first_tdp);
	first_tdp->prev = last_tdp;
	pChan->dq_tx_last = first_tdp;		/* last descriptor to be transmitted */
	pChan->dq_tx_cleanup = first_tdp;	/* first descriptor to be cleaned up */

	return STATUS_SUCCESS;
}

/*
 * initialize chip, called during driver initialization procedure
 * action sequency was carefully chosen, don't mess with it
 */
NTSTATUS pciscc_chip_open(
	IN PLOCAL_DEVICE_INFO pLDI
	)
{
	LARGE_INTEGER	time;
	LARGE_INTEGER	time_expire;

	if (pLDI->initialized)
		return STATUS_SUCCESS;
	/* allocate DMS buffers */
	pLDI->DmaBuffers = pciscc_alloc(sizeof(DMA_BUFFERS));
	if (pLDI->DmaBuffers == NULL)
		return STATUS_NO_MEMORY;
	/* peripheral interrupt queue */
	pLDI->iq_per = pLDI->DmaBuffers->iq_per;
	/* configuration interrupt queue */
	pLDI->iq_cfg = pLDI->DmaBuffers->iq_cfg;
	/* Map base address */
	pLDI->MappedAddress = MmMapIoSpace(pLDI->Address,pLDI->AddrLen,FALSE);
	/* Plug in ISR function */
	IoConnectInterrupt(	&pLDI->IrqObject, pciscc_isr, pLDI,	NULL,
						pLDI->IrqVector, pLDI->IrqLevel, pLDI->IrqLevel,
						LevelSensitive,	TRUE, pLDI->IrqAffinity, FALSE );
	/* Init queue pointer */
	RtlZeroMemory(pLDI->iq_per, 4*CFG_IQLEN);
	RtlZeroMemory(pLDI->iq_cfg, 4*CFG_IQLEN);
	pLDI->iq_per_next = pLDI->iq_per;
	pLDI->iq_cfg_next = pLDI->iq_cfg;
	/* global hardware initialization */
	WriteL(pLDI,GMODE, (4 * PERCFG) | (3 * LCD) | CMODE);
	WriteL(pLDI,IQPBAR,VirtToPhys(pLDI->iq_per));
	WriteL(pLDI,IQCFGBAR,VirtToPhys(pLDI->iq_cfg));
	WriteL(pLDI,IQLENR2,(((CFG_IQLEN/32)-1)*IQCFGLEN) | (((CFG_IQLEN/32)-1)*IQPLEN));
	WriteL(pLDI,FIFOCR1,((32/4)*TFSIZE0)
		| ((32/4)*TFSIZE1)
		| ((32/4)*TFSIZE2)
		| ((32/4)*TFSIZE3));
	WriteL(pLDI,FIFOCR2,((24/4)*TFRTHRES0)
		| ((24/4)*TFRTHRES1)
		| ((24/4)*TFRTHRES2)
		| ((24/4)*TFRTHRES3)
		| M4_0 | M4_1 | M4_2 | M4_3);
	WriteL(pLDI,FIFOCR3,CFG_FIFO_RX_T);
	WriteL(pLDI,FIFOCR4,(20*TFFTHRES0)
		| (20*TFFTHRES1)
		| (20*TFFTHRES2)
		| (20*TFFTHRES3));
	/* mask out all DMAC interrupts */
	WriteL(pLDI,CH0CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH1CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH2CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH3CFG,(MRFI | MTFI | MRERR | MTERR));
	/* all SCC cores in reset state */
	WriteL(pLDI,SCCBASE[0]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[1]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[2]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[3]+CCR0,0x00000000);
	/* mask out all SCC interrupts */
	WriteL(pLDI,SCCBASE[0]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[1]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[2]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[3]+IMR,0xffffffff);
	/* peripheral configuration */
	WriteL(pLDI,LCONF,(BTYP*3));
	WriteL(pLDI,SSCCON,0x00000000);
	WriteL(pLDI,SSCIM,0x00000000);
	WriteL(pLDI,GPDIR,0x000000ff);
	WriteL(pLDI,GPDATA,0x00000000);
	WriteL(pLDI,GPIM,0x00000000);
	/* initialize configuration and peripheral IQs */
	pLDI->mailbox = MAILBOX_NONE;
	WriteL(pLDI,GCMDR, CFGIQCFG | CFGIQP | AR);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pLDI->mailbox) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if (pLDI->mailbox == MAILBOX_OK) {
		pLDI->initialized = 1;
		return STATUS_SUCCESS;
	}
	/* global configuration to reset state */
	WriteL(pLDI,GMODE,(4 * PERCFG) | (3 * LCD) | CMODE | OSCPD);
	/* mask all DMAC interrupts */
	WriteL(pLDI,CH0CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH1CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH2CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH3CFG,(MRFI | MTFI | MRERR | MTERR));
	/* SCC cores to reset state */
	WriteL(pLDI,SCCBASE[0]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[1]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[2]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[3]+CCR0,0x00000000);
	/* mask all SCC interrupts */
	WriteL(pLDI,SCCBASE[0]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[1]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[2]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[3]+IMR,0xffffffff);
	/* queues no longer exist */
	pLDI->iq_per_next = NULL;
	pLDI->iq_cfg_next = NULL;
	/* Shut down ISR routine */
	IoDisconnectInterrupt(pLDI->IrqObject);
	/* Unmap memory */
	MmUnmapIoSpace(pLDI->MappedAddress,pLDI->AddrLen);
	/* Release buffers */
	pciscc_free(pLDI->DmaBuffers);
	/* Print debug message */
	DbgPrint("PCISCC: Interrupt queue error (pciscc_chip_open)\n");
	return STATUS_IO_TIMEOUT;
}

/*
 * close chip, called when last device (channel) of a chip was closed.
 * don't mess up.
 */
VOID pciscc_chip_close(
	IN PLOCAL_DEVICE_INFO pLDI
	)
{
	ULONG chan;

	if (!pLDI->initialized)
		return;
	/* close channels */
	for (chan = 0; chan < 4; chan++)
		pciscc_channel_close(pLDI,chan);
	pLDI->initialized = 0;
	/* global configuration to reset state */
	WriteL(pLDI,GMODE,(4 * PERCFG) | (3 * LCD) | CMODE | OSCPD);
	/* mask all DMAC interrupts */
	WriteL(pLDI,CH0CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH1CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH2CFG,(MRFI | MTFI | MRERR | MTERR));
	WriteL(pLDI,CH3CFG,(MRFI | MTFI | MRERR | MTERR));
	/* SCC cores to reset state */
	WriteL(pLDI,SCCBASE[0]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[1]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[2]+CCR0,0x00000000);
	WriteL(pLDI,SCCBASE[3]+CCR0,0x00000000);
	/* mask all SCC interrupts */
	WriteL(pLDI,SCCBASE[0]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[1]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[2]+IMR,0xffffffff);
	WriteL(pLDI,SCCBASE[3]+IMR,0xffffffff);
	/* queues no longer exist */
	pLDI->iq_per_next = NULL;
	pLDI->iq_cfg_next = NULL;
	/* Shut down ISR routine */
	IoDisconnectInterrupt(pLDI->IrqObject);
	/* Unmap memory */
	MmUnmapIoSpace(pLDI->MappedAddress,pLDI->AddrLen);
	/* Release buffers */
	pciscc_free(pLDI->DmaBuffers);
	/* Realease pending RX frames */
	pciscc_clear_rxpending(pLDI);
	pciscc_clear_rxlist(pLDI);
}

/*
 * open one channel, the sequence of actions done here was carefully chosen,
 * don't mess with it unless you know exactly what you are doing...
 */
NTSTATUS pciscc_channel_open(
	PLOCAL_DEVICE_INFO pLDI,
	ULONG chan,
	ULONG bitrate,
	ULONG mode
	)
{
	PCHANNEL_INFO pChan;
	NTSTATUS Status;
	LARGE_INTEGER	time;
	LARGE_INTEGER	time_expire;
	ULONG l;

	if (!pLDI->initialized)
		return STATUS_NO_SUCH_DEVICE;
	if (!bitrate)
		return STATUS_INVALID_PARAMETER;
	pChan = &pLDI->Channel[chan];
	/* nothing to change */
	if ((bitrate == pChan->bitrate) && (mode == pChan->mode))
		return STATUS_SUCCESS;
	/* reset channel configuration */
	pciscc_channel_close(pLDI,chan);
	/* initialize DMA queues */
	Status = pciscc_init_queues(pLDI,chan);
	if (!NT_SUCCESS(Status))
		return Status;
	/* set bitrate/mode */
	pChan->mode = mode;
	pChan->bitrate = bitrate;
	/* stop SCC */
	WriteL(pLDI,SCCBASE[chan]+CCR1,0);
	WriteL(pLDI,SCCBASE[chan]+CCR2,0);
	WriteL(pLDI,SCCBASE[chan]+CCR0,0);
	pChan->ccr0 = pChan->ccr1 = pChan->ccr2 = 0;
	WriteL(pLDI,SCCBASE[chan]+TIMR,0);
	/* set IQ lengths and base addresses */
	l = ReadL(pLDI,IQLENR1);
	switch (chan) {
	case 0:	WriteL(pLDI,CH0CFG,MRFI | MTFI | MRERR | MTERR);
			WriteL(pLDI,IQSCC0RXBAR,VirtToPhys(pChan->iq_rx));
			WriteL(pLDI,IQSCC0TXBAR,VirtToPhys(pChan->iq_tx));
			l &= 0x0fff0fff;
			l |= (((CFG_IQLEN/32)-1)*IQSCC0RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC0TXLEN);
			break;
	case 1:	WriteL(pLDI,CH1CFG,MRFI | MTFI | MRERR | MTERR);
			WriteL(pLDI,IQSCC1RXBAR,VirtToPhys(pChan->iq_rx));
			WriteL(pLDI,IQSCC1TXBAR,VirtToPhys(pChan->iq_tx));
			l &= 0xf0fff0ff;
			l |= (((CFG_IQLEN/32)-1)*IQSCC1RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC1TXLEN);
			break;
	case 2:	WriteL(pLDI,CH2CFG,MRFI | MTFI | MRERR | MTERR);
			WriteL(pLDI,IQSCC2RXBAR,VirtToPhys(pChan->iq_rx));
			WriteL(pLDI,IQSCC2TXBAR,VirtToPhys(pChan->iq_tx));
			l &= 0xff0fff0f;
			l |= (((CFG_IQLEN/32)-1)*IQSCC2RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC2TXLEN);
			break;
	case 3:	WriteL(pLDI,CH3CFG,MRFI | MTFI | MRERR | MTERR);
			WriteL(pLDI,IQSCC3RXBAR,VirtToPhys(pChan->iq_rx));
			WriteL(pLDI,IQSCC3TXBAR,VirtToPhys(pChan->iq_tx));
			l &= 0xfff0fff0;
			l |= (((CFG_IQLEN/32)-1)*IQSCC3RXLEN) | (((CFG_IQLEN/32)-1)*IQSCC3TXLEN);
			break;
	}
	WriteL(pLDI,IQLENR1,l);
	pLDI->mailbox = MAILBOX_NONE;
	WriteL(pLDI,GCMDR, AR	/* Action Request */
			| (chan == 0 ? (CFGIQSCC0RX | CFGIQSCC0TX) : 0)
			| (chan == 1 ? (CFGIQSCC1RX | CFGIQSCC1TX) : 0)
			| (chan == 2 ? (CFGIQSCC2RX | CFGIQSCC2TX) : 0)
			| (chan == 3 ? (CFGIQSCC3RX | CFGIQSCC3TX) : 0));
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pLDI->mailbox) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if (pLDI->mailbox != MAILBOX_OK) {
		DbgPrint("PCISCC: TX/RX interrupt queue error (pciscc_channel_open)\n");
		pChan->stats.io_error++;
	}
	/* initialize channel's SCC core */
	pChan->ccr0 = PU | VIS | PSD;
	pChan->ccr0 |= (mode & MODE_z) ? 0 : 2*SC;
	pChan->ccr0 |= (mode & MODE_r) ? 0 : 6*CM;
	pChan->ccr0 |= (mode & MODE_t) ? 0 : SSEL;
	pChan->ccr1 = ODS | ICD | FCTS | MDS1 | SFLAG;
	pChan->ccr2 = RAC | (3*RFTH) | ITF;
	WriteL(pLDI,SCCBASE[chan]+CCR0,pChan->ccr0);
	WriteL(pLDI,SCCBASE[chan]+CCR1,pChan->ccr1);
	WriteL(pLDI,SCCBASE[chan]+CCR2,pChan->ccr2);
	pciscc_set_baud(pLDI,chan,(mode & MODE_r) ? bitrate : 16*bitrate);
	WriteL(pLDI,SCCBASE[chan]+RLCR,RCE | (CFG_MTU*RL));
	/*
	 * all sent | tx device underrun | timer int | tx message repeat |
	 * tx pool ready | rx device overflow | receive FIFO overflow |
	 * frame length exceeded => interrupt mask register
	 */
	WriteL(pLDI,SCCBASE[chan]+IMR,~(ALLS | XDU | TIN | XMR | XPR | RDO | RFO | FLEX));
	/* wait until command executing (CEC) is clear */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		l = ReadL(pLDI,SCCBASE[chan]+STAR);
		if (!(l & CEC)) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	if (l & CEC)
		DbgPrint("PCISCC: Timeout waiting for SCC being ready for reset.\n");
	/* RX and TX SCC reset */
	pChan->tx_mailbox = 0xffffffff;
	WriteL(pLDI,SCCBASE[chan]+CMDR,RRES | XRES);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pChan->tx_mailbox!=0xffffffff) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if ((pChan->tx_mailbox & 0x03ffffff) != 0x02001000) {
		DbgPrint("PCISCC: TX/RX reset error (pciscc_channel_open)\n");
		pChan->stats.io_error++;
	}
	/* initialize DMAC channel's RX */
	switch (chan) {
	case 0:	WriteL(pLDI,CH0CFG,IDR);
			WriteL(pLDI,CH0BRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH0FRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH0LRDA,VirtToPhys(pChan->dq_rx_next->prev->prev));
			break;
	case 1:	WriteL(pLDI,CH1CFG,IDR);
			WriteL(pLDI,CH1BRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH1FRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH1LRDA,VirtToPhys(pChan->dq_rx_next->prev->prev));
			break;
	case 2:	WriteL(pLDI,CH2CFG,IDR);
			WriteL(pLDI,CH2BRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH2FRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH2LRDA,VirtToPhys(pChan->dq_rx_next->prev->prev));
			break;
	case 3:	WriteL(pLDI,CH3CFG,IDR);
			WriteL(pLDI,CH3BRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH3FRDA,VirtToPhys(pChan->dq_rx));
			WriteL(pLDI,CH3LRDA,VirtToPhys(pChan->dq_rx_next->prev->prev));
			break;
	}
	pLDI->mailbox = MAILBOX_NONE;
	WriteL(pLDI,GCMDR,AR);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pLDI->mailbox) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if (pLDI->mailbox != MAILBOX_OK) {
		DbgPrint("PCISCC: RX DMAC config error (pciscc_channel_open)\n");
		pChan->stats.io_error++;
	}
	/* mask all DMAC interrupts (needed) */
	switch (chan) {
	case 0:	WriteL(pLDI,CH0CFG,MRFI | MTFI | MRERR | MTERR);
			break;
	case 1:	WriteL(pLDI,CH1CFG,MRFI | MTFI | MRERR | MTERR);
			break;
	case 2:	WriteL(pLDI,CH2CFG,MRFI | MTFI | MRERR | MTERR);
			break;
	case 3:	WriteL(pLDI,CH3CFG,MRFI | MTFI | MRERR | MTERR);
			break;
	}
	/* wait until command executing (CEC) is clear */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		l = ReadL(pLDI,SCCBASE[chan]+STAR);
		if (!(l & CEC)) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	if (l & CEC)
		DbgPrint("PCISCC: Timeout waiting for SCC being ready for reset.\n");
	/* SCC core TX reset (again) */
	pChan->tx_mailbox = 0xffffffff;
	WriteL(pLDI,SCCBASE[chan]+CMDR,XRES);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pChan->tx_mailbox!=0xffffffff) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if ((pChan->tx_mailbox & 0x03ffffff) != 0x02001000) {
		DbgPrint("PCISCC: TX reset error (pciscc_channel_open)\n");
		pChan->stats.io_error++;
	}
	/*
	 * initialize DMAC's TX channel, FI must stay masked all the time
	 * even during operation, see device errata 03/99
	 */
	switch (chan) {
	case 0:	WriteL(pLDI,CH0CFG,IDT | MTFI);
			WriteL(pLDI,CH0BTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH0FTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH0LTDA,VirtToPhys(pChan->dq_tx));
			break;
	case 1:	WriteL(pLDI,CH1CFG,IDT | MTFI);
			WriteL(pLDI,CH1BTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH1FTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH1LTDA,VirtToPhys(pChan->dq_tx));
			break;
	case 2:	WriteL(pLDI,CH2CFG,IDT | MTFI);
			WriteL(pLDI,CH2BTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH2FTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH2LTDA,VirtToPhys(pChan->dq_tx));
			break;
	case 3:	WriteL(pLDI,CH3CFG,IDT | MTFI);
			WriteL(pLDI,CH3BTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH3FTDA,VirtToPhys(pChan->dq_tx));
			WriteL(pLDI,CH3LTDA,VirtToPhys(pChan->dq_tx));
			break;
	}
	pLDI->mailbox = MAILBOX_NONE;
	WriteL(pLDI,CMDR,AR);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pLDI->mailbox) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if (pLDI->mailbox != MAILBOX_OK) {
		DbgPrint("PCISCC: TX DMAC config error (pciscc_channel_open)\n");
		pChan->stats.io_error++;
	} else
		pChan->txstate=TX_IDLE;
	/* initialization complete */
	pChan->initialized = 1;
	return STATUS_SUCCESS;
}

/* close one channel - don't mess with it either */
VOID pciscc_channel_close(
	PLOCAL_DEVICE_INFO pLDI,
	ULONG chan
	)
{
	PCHANNEL_INFO pChan;
	struct rx_desc *rdp;
	LARGE_INTEGER	time;
	LARGE_INTEGER	time_expire;
	ULONG i,l;

	pChan = &pLDI->Channel[chan];
	/* check init state */
	if (!pChan->initialized)
		return;
	pChan->initialized = 0;
	/* at first stop timer */
	WriteL(pLDI,SCCBASE[chan]+TIMR,0);
	/* wait until command executing (CEC) is clear */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		l = ReadL(pLDI,SCCBASE[chan]+STAR);
		if (!(l & CEC)) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	if (l & CEC)
		DbgPrint("PCISCC: Timeout waiting for SCC being ready for reset.\n");
	/* RX and TX SCC reset */
	pChan->tx_mailbox = 0xffffffff;
	WriteL(pLDI,SCCBASE[chan]+CMDR,RRES | XRES);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pChan->tx_mailbox!=0xffffffff) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if ((pChan->tx_mailbox & 0x03ffffff) != 0x02001000) {
		DbgPrint("PCISCC: TX/RX reset error (pciscc_channel_close)\n");
		pChan->stats.io_error++;
	}
	pChan->txstate = TX_RESET;
	/* stop SCC core */
	WriteL(pLDI,SCCBASE[chan]+CCR1,0);
	WriteL(pLDI,SCCBASE[chan]+CCR2,0);
	WriteL(pLDI,SCCBASE[chan]+CCR0,0);
	pChan->ccr0 = pChan->ccr1 = pChan->ccr2 = 0;
	pChan->bitrate = pChan->mode = 0;
	/*
	 * Give the isr some time to "refill" the rx-dq
	 * we _MUST_ guarantee that the DMAC-RX is _NOT_ in
	 * hold state when issuing the RESET command, otherwise the DMAC
	 * will crash. (DSCC-4 Rev. <= 2.1)
	 * In addition to that we may only issue a RESET if channel is
	 * currently really initialized, otherwise something horrible will
	 * result.
	 */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* OK, now we should be ready to put the DMAC into reset state */
	switch (chan) {
	case 0:	WriteL(pLDI,CH0CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
			break;
	case 1:	WriteL(pLDI,CH1CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
			break;
	case 2:	WriteL(pLDI,CH2CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
			break;
	case 3: WriteL(pLDI,CH3CFG,RDR | RDT | MRFI | MTFI | MRERR | MTERR);
			break;
	}
	pLDI->mailbox = MAILBOX_NONE;
	WriteL(pLDI,GCMDR,AR);
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pLDI->mailbox) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if (pLDI->mailbox != MAILBOX_OK) {
		DbgPrint("PCISCC: DMAC reset error (pciscc_channel_close)\n");
		pChan->stats.io_error++;
	}
	/* clear IQs */
	l = ReadL(pLDI,IQLENR1);
	switch (chan) {
	case 0:	l &= 0x0fff0fff;
			WriteL(pLDI,IQLENR1,l);
			WriteL(pLDI,IQSCC0RXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			WriteL(pLDI,IQSCC0TXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			break;
	case 1: l &= 0xf0fff0ff;
			WriteL(pLDI,IQLENR1,l);
			WriteL(pLDI,IQSCC1RXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			WriteL(pLDI,IQSCC1TXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			break;
	case 2: l &= 0xff0fff0f;
			WriteL(pLDI,IQLENR1,l);
			WriteL(pLDI,IQSCC2RXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			WriteL(pLDI,IQSCC2TXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			break;
	case 3: l &= 0xfff0fff0;
			WriteL(pLDI,IQLENR1,l);
			WriteL(pLDI,IQSCC3RXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			WriteL(pLDI,IQSCC3TXBAR,VirtToPhys(pLDI->DmaBuffers->dummy));
			break;
	}
	pLDI->mailbox = MAILBOX_NONE;
	WriteL(pLDI,GCMDR,AR
			| (chan == 0 ? (CFGIQSCC0RX | CFGIQSCC0TX) : 0)
			| (chan == 1 ? (CFGIQSCC1RX | CFGIQSCC1TX) : 0)
			| (chan == 2 ? (CFGIQSCC2RX | CFGIQSCC2TX) : 0)
			| (chan == 3 ? (CFGIQSCC3RX | CFGIQSCC3TX) : 0));
	/* wait for interrupt handler */
	KeQuerySystemTime(&time);
	time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
	do {
		if (pLDI->mailbox) break;
		KeQuerySystemTime(&time);
	} while RtlLargeIntegerLessThan(time,time_expire);
	/* mailbox was written by isr */
	if (pLDI->mailbox != MAILBOX_OK) {
		DbgPrint("PCISCC: Interrupt queue reset error (pciscc_channel_close)");
		pChan->stats.io_error++;
	}
	/* queues no longer exist */
	pChan->iq_rx_next=NULL;
	pChan->iq_tx_next=NULL;
	/* release RX frames from queue */
	for (i=0; i < CFG_RX_DESC; i++)
	{
		if (pChan->dq_rx[i].l1_frame)
			pciscc_free(pChan->dq_rx[i].l1_frame);
	}
}

/* find best BRG N and M match for given bitrate */
VOID pciscc_set_baud(
	PLOCAL_DEVICE_INFO pLDI,
	ULONG chan,
	ULONG rate
	)
{
	LONG brg_best_n	= 0;
	LONG brg_best_m	= 0;
	LONG brg_tmp,m,n;
	ULONG brg_error	= XTAL_FREQ;
	ULONG brg_tmp_error;

	for (m=0; m<16; m++) {
		for (n=0; n<63; n++) {
			brg_tmp = (n+1)*(1<<m);
			brg_tmp_error = labs(XTAL_FREQ/brg_tmp - rate);
			if (brg_tmp_error < brg_error) {
				brg_best_m = m;
				brg_best_n = n;
				brg_error = brg_tmp_error;
			}
		}
	}
	DbgPrint("PCISCC: set BRR%d M=%d, N=%d\n",chan,brg_best_m,brg_best_n);
	WriteL(pLDI,SCCBASE[chan]+BRR,(brg_best_m*BRM) | (brg_best_n*BRN));
}

/* report channel state (PTT/DCD) */
ULONG pciscc_state(
	IN PLOCAL_DEVICE_INFO pLDI,
	IN ULONG chan
	)
{
	PCHANNEL_INFO pChan;
	ULONG l;
	ULONG state = 0;

	pChan = &pLDI->Channel[chan];
	if (!pChan->initialized)
		return CH_DEAD;
	ReadL(pLDI,SCCBASE[chan]+STAR);
	l = ReadL(pLDI,SCCBASE[chan]+STAR);
	if (!(l & CD))
		state |= CH_DCD;
	if (!IsListEmpty(&pLDI->RxQueue))
		state |= CH_RXB;
	switch(pChan->txstate) {
	case TX_DELAY:
	case TX_XMIT:
	case TX_TAIL:
		state |= CH_PTT;
		break;
	case TX_CAL:
		state |= CH_TBY | CH_PTT;
		break;
	case TX_RESET:
		state |= CH_TBY;
		break;
	}
	return state;
}

/* send calibration signal (IRQ synchronized) */
BOOLEAN pciscc_calib_sync(
	IN PCALIB_SYNC_INFO pInfo
	)
{
	PLOCAL_DEVICE_INFO pLDI = pInfo->pLDI;
	ULONG chan = pInfo->chan;
	PCHANNEL_INFO pChan = &pLDI->Channel[chan];
	ULONG bits = pInfo->time * pChan->bitrate * 60;

	if (pChan->initialized) {
		if ((pChan->txstate == TX_CAL) && !bits) {
			WriteL(pLDI,SCCBASE[chan]+TIMR,MIN_FLAGS*TVALUE);
			WriteL(pLDI,SCCBASE[chan]+CMDR,STI);
		}
		if ((pChan->txstate != TX_RESET) && (pChan->txstate != TX_STALL) && bits) {
			pChan->ccr1 |= RTS;
			WriteL(pLDI,SCCBASE[chan]+CCR1,pChan->ccr1);
			WriteL(pLDI,SCCBASE[chan]+TIMR,bits*TVALUE);
			WriteL(pLDI,SCCBASE[chan]+CMDR,STI);
			pChan->txstate = TX_CAL;
		}
	}
	return TRUE;
}

/* XMIT frame (IRQ synchronized) */
BOOLEAN pciscc_xmit_sync(
	IN PXMIT_SYNC_INFO pInfo
	)
{
	PLOCAL_DEVICE_INFO pLDI = pInfo->pLDI;
	L1FRAME *tx_frame = pInfo->tx_frame;
	BOOLEAN enqueued = FALSE;
	PCHANNEL_INFO pChan;
	struct tx_desc *tdp;
	ULONG chan;
	ULONG txd_bits;
	
	/* get channel number */
	chan = tx_frame->chan;
	if (chan < 0 || chan > 3)
		return FALSE;
	pChan = &pLDI->Channel[chan];
	if (!pChan->initialized)
		return FALSE;
	/* descriptor list already full ? */
	tdp = pChan->dq_tx_last->next;
	if (tdp->l1_frame == NULL && tdp->next->l1_frame == NULL) {
		/* insert TX frame buffer into descriptor queue */
		if (pChan->txstate == TX_IDLE ||
			pChan->txstate == TX_DELAY ||
			pChan->txstate == TX_XMIT ||
			pChan->txstate == TX_TAIL)
		{
			tdp->l1_frame = tx_frame;
			tdp->dataptr = VirtToPhys(tx_frame->frame);
			tdp->result = 0;
			tdp->flags = FE | (NO * (tx_frame->len & 0x0fff));
			pChan->dq_tx_last = tdp;
			pChan->tx_enqueued++;
			enqueued = TRUE;
		}
		/* calculate tx keyup delay */
		txd_bits = pChan->bitrate * tx_frame->txdelay / 100;
		if (txd_bits < MIN_FLAGS) txd_bits = MIN_FLAGS;
		/* TX was idle, key up and start txdelay */
		if (pChan->txstate == TX_IDLE) {
			pChan->txstate=TX_DELAY;
			pChan->ccr1 |= RTS;
			WriteL(pLDI,SCCBASE[chan]+CCR1,pChan->ccr1);
			WriteL(pLDI,SCCBASE[chan]+TIMR,0);
			WriteL(pLDI,SCCBASE[chan]+TIMR,txd_bits*TVALUE);
			WriteL(pLDI,SCCBASE[chan]+CMDR,STI);
		}
		/* Update TX last descpriptor address if we are in TX_XMIT state */
		if (pChan->txstate == TX_XMIT) {
			switch (chan) {
			case 0:	WriteL(pLDI,CH0LTDA,VirtToPhys(tdp));
					break;
			case 1:	WriteL(pLDI,CH1LTDA,VirtToPhys(tdp));
					break;
			case 2:	WriteL(pLDI,CH2LTDA,VirtToPhys(tdp));
					break;
			case 3:	WriteL(pLDI,CH3LTDA,VirtToPhys(tdp));
					break;
			}
		}
	}
	return enqueued;
}

/* ---------------------------------------------------------------- */

/* interrupt service routine */
BOOLEAN pciscc_isr(
	IN PKINTERRUPT Interrupt,
	IN PVOID ServiceContext
	)
{
	PLOCAL_DEVICE_INFO pLDI = (PLOCAL_DEVICE_INFO)ServiceContext;
	PCHANNEL_INFO pChan;
	ULONG status;
	ULONG chan;
	PULONG iqp;

	/* ack' irq */
	status = ReadL(pLDI,GSTAR);
	WriteL(pLDI,GSTAR,status);
	/* process peripheral queue */
	iqp = pLDI->iq_per_next;
	while (iqp && *iqp) {
		*(iqp++)=0;
		if (iqp == pLDI->iq_per+CFG_IQLEN)
			iqp = pLDI->iq_per;
	}
	pLDI->iq_per_next = iqp;
	/* process configuration queue */
	iqp = pLDI->iq_cfg_next;
	while(iqp && *iqp) {
		if ((*iqp) & ARACK)
			pLDI->mailbox = MAILBOX_OK;
		if ((*iqp) & ARF)
			pLDI->mailbox = MAILBOX_FAILURE;
		*(iqp++)=0;
		if (iqp == pLDI->iq_cfg+CFG_IQLEN)
			iqp = pLDI->iq_cfg;
	}
	pLDI->iq_cfg_next = iqp;
	/* process channel interrupt queues */
	for (chan = 0; chan < 4; chan++) {
		pChan = &pLDI->Channel[chan];
		/* process RX queue */
		iqp = pChan->iq_rx_next;
		while(iqp && *iqp) {
			if (pChan->initialized) {
				/* statistics only */
				if ((*iqp & SCCIV_SCC) && (*iqp & SCCIV_RDO))
					pChan->stats.rx_bufferoverflow++;
				if ((*iqp & SCCIV_SCC) && (*iqp & SCCIV_RFO))
					pChan->stats.rx_bufferoverflow++;
				if ((*iqp & SCCIV_SCC) && (*iqp & SCCIV_FLEX))
					pChan->stats.rx_bufferoverflow++;
			}
			pChan->rx_mailbox = *iqp;
			*(iqp++)=0;
			if (iqp == pChan->iq_rx+CFG_IQLEN)
				iqp = pChan->iq_rx;
		}
		pChan->iq_rx_next = iqp;
		/* process TX queue */
		iqp = pChan->iq_tx_next;
		while (iqp && *iqp) {
			if (pChan->initialized) {
				if (*iqp & TIN) {
					/* timer interrupt */
					WriteL(pLDI,SCCBASE[chan]+TIMR,0);
					if (pChan->tx_enqueued) {
						pChan->txstate = TX_XMIT;
						switch (chan) {
						case 0:	WriteL(pLDI,CH0LTDA,VirtToPhys(pChan->dq_tx_last));
								break;
						case 1:	WriteL(pLDI,CH1LTDA,VirtToPhys(pChan->dq_tx_last));
								break;
						case 2:	WriteL(pLDI,CH2LTDA,VirtToPhys(pChan->dq_tx_last));
								break;
						case 3:	WriteL(pLDI,CH3LTDA,VirtToPhys(pChan->dq_tx_last));
								break;
						}
					} else {
						pChan->txstate=TX_IDLE;
						pChan->ccr1 &= ~RTS;
						WriteL(pLDI,SCCBASE[chan]+CCR1,pChan->ccr1);
					}
				}
				if (*iqp & ALLS) {
					/* complete TX-frame sent out */
					if (pChan->tx_enqueued)
						pChan->tx_enqueued--;
					if (!pChan->tx_enqueued) {
						pChan->txstate=TX_TAIL;
						WriteL(pLDI,SCCBASE[chan]+TIMR,MIN_FLAGS*TVALUE);
						WriteL(pLDI,SCCBASE[chan]+CMDR,STI);
					}
				}
				if (*iqp & XDU) {
					/* TX data underrun */
					KeInsertQueueDpc(&pLDI->TxResetDpc,(PVOID)chan,NULL);
				}
			}
			pChan->tx_mailbox = *iqp;
			*(iqp++)=0;
			if (iqp == pChan->iq_tx+CFG_IQLEN)
				iqp = pChan->iq_tx;
		}
		pChan->iq_tx_next = iqp;
	}
	/* start bottom half DPC functions */
	if (status) {
		KeInsertQueueDpc(&pLDI->IsrDpc,NULL,NULL);
		return TRUE;
	}
	return FALSE;
}

/* ---------------------------------------------------------------- */

/* TX underrun reset DPC */
VOID pciscc_dpc_txreset(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT pDO,
	IN PVOID SystemArgument1,
	IN PVOID SystemArgument2
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	int chan = (int)SystemArgument1;
	PCHANNEL_INFO pChan = &pLDI->Channel[chan];
	LARGE_INTEGER	time;
	LARGE_INTEGER	time_expire;

	KeAcquireSpinLockAtDpcLevel(&pLDI->Lock);
	if (pChan->initialized) {
		pChan->txstate=TX_STALL;
		pChan->stats.tx_error++;
	
		pChan->ccr1 &= ~RTS;
		WriteL(pLDI,SCCBASE[chan]+CCR1,pChan->ccr1);
		WriteL(pLDI,SCCBASE[chan]+CMDR,XRES);

		KeQuerySystemTime(&time);
		time_expire = RtlLargeIntegerAdd(time,RtlConvertUlongToLargeInteger(CHIP_TIMEOUT));
		do {
			KeQuerySystemTime(&time);
		} while RtlLargeIntegerLessThan(time,time_expire);

		pChan->txstate = TX_IDLE;
	}
	KeReleaseSpinLockFromDpcLevel(&pLDI->Lock);
}

/* ---------------------------------------------------------------- */

/* TX/RX queue processing DPC routine */
VOID pciscc_dpc_isr(
	IN PKDPC Dpc,
	IN PDEVICE_OBJECT pDO,
	IN PVOID SystemArgument1,
	IN PVOID SystemArgument2
	)
{
	PLOCAL_DEVICE_INFO pLDI = pDO->DeviceExtension;
	PCHANNEL_INFO pChan;
	struct rx_desc *rdp;
	struct tx_desc *tdp;
	L1FRAME *rx_new;
	USHORT bno;
	UCHAR rdsb;		/* receive data status byte */
	ULONG chan,status,valid,rx_cnt;

	KeAcquireSpinLockAtDpcLevel(&pLDI->Lock);
	for (chan = 0; chan < 4; chan++) {
		pChan = &pLDI->Channel[chan];
		if (!pChan->initialized)
			continue;
		/* transmit DMA queue */
		tdp = pChan->dq_tx_cleanup;
		while (tdp->result & C) {
			/* clean up all (C)omplete descriptors */
			if (tdp->l1_frame)
				pciscc_free(tdp->l1_frame);
			tdp->l1_frame = NULL;
			tdp->dataptr = VirtToPhys(pLDI->DmaBuffers->dummy);
			tdp->flags = FE | (8*NO);
			tdp->result = 0;
			tdp = tdp->next;
			pChan->stats.tx_frames++;
		}
		pChan->dq_tx_cleanup = tdp;
		/* receiver DMA queue */
		rdp = pChan->dq_rx_next;
		rx_cnt = 0;
		while (rdp->result & C) {
			if ((rdp->nextptr != VirtToPhys(rdp->next)) || (rdp->dataptr != VirtToPhys(rdp->l1_frame->frame))) {
				pChan->stats.rx_bufferoverflow++;
			}
			status = rdp->result;
			bno = (USHORT)(status >> 16) & 0x1fff;
			valid = 1;					/* we assume frame valid */
			if ((status & RA) || (bno < 5) || (bno > CFG_MTU) || !(status & FE) || (rdp->feptr != VirtToPhys(rdp))) {
				valid = 0;				/* aborted or invalid length */
			} else {
				rdsb = rdp->l1_frame->frame[bno-1];
				if (!(rdsb & SB_VFR))	/* incorrect bit length */
					valid = 0;
				if (rdsb & SB_RDO)		/* data overflow */
					valid = 0;
				if (!(rdsb & SB_CRC))	/* CRC error */
					valid = 0;
				if (rdsb & SB_RAB)		/* receive message aborted */
					valid = 0;
			}
			/* OK, this is a little bit tricky. We have to make sure
			 * that every descriptor has a buffer assigned. Thus we
			 * can only release a buffer to the link layer if we get
			 * a new one in turn before. */
			if (valid && pLDI->OpenCnt) {
				if ((rx_new = pciscc_alloc(sizeof(L1FRAME))) != NULL) {
					rdp->l1_frame->len = bno-1;
					rdp->l1_frame->chan = (UCHAR)chan;
					rdp->l1_frame->txdelay = 0;
					InsertTailList(&pLDI->RxQueue, &rdp->l1_frame->list_entry);
					rdp->l1_frame = rx_new;
					rdp->dataptr = VirtToPhys(rx_new->frame);
					pChan->stats.rx_frames++;
				} else
					pChan->stats.rx_overrun++;
			}
			rdp->flags = CFG_MTU*NO;
			rdp->result = 0;
			rdp->feptr = 0;
			rdp = rdp->next;
			rx_cnt++;
		}
		pChan->dq_rx_next = rdp;
		/*
		 * tell DMAC last available descriptor - keep up one
		 * descriptor space for security (paranoia) (...->prev->prev)
		 */
		if (rx_cnt) {
			switch (chan) {
			case 0:	WriteL(pLDI,CH0LRDA,VirtToPhys(rdp->prev->prev));
					break;
			case 1:	WriteL(pLDI,CH1LRDA,VirtToPhys(rdp->prev->prev));
					break;
			case 2:	WriteL(pLDI,CH2LRDA,VirtToPhys(rdp->prev->prev));
					break;
			case 3:	WriteL(pLDI,CH3LRDA,VirtToPhys(rdp->prev->prev));
					break;
			}
		}
	}
	KeReleaseSpinLockFromDpcLevel(&pLDI->Lock);

	/* Process pending IRPs */
	KeAcquireSpinLockAtDpcLevel(&pLDI->Lock);
    while (!IsListEmpty(&pLDI->RxPending) && !IsListEmpty(&pLDI->RxQueue))
	{
		PLIST_ENTRY pEntry;
		PIO_STACK_LOCATION pIrpStack;
		PIRP pIrp;
		L1FRAME *rx_frame;
		ULONG Count = 0;

        pEntry = RemoveHeadList(&pLDI->RxPending);
		pIrp = CONTAINING_RECORD(pEntry,IRP,Tail.Overlay.ListEntry);

		/* Get read buffer length */
		pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
		Count = pIrpStack->Parameters.Read.Length;
		if (Count > sizeof(L1FRAME))
			Count = sizeof(L1FRAME);

		/* Get RX frame */
		pEntry = RemoveHeadList(&pLDI->RxQueue);
		rx_frame = CONTAINING_RECORD(pEntry, L1FRAME, list_entry);
		RtlCopyMemory(pIrp->AssociatedIrp.SystemBuffer, rx_frame, Count);
		pciscc_free(rx_frame);

		pIrp->IoStatus.Status = STATUS_SUCCESS;
		pIrp->IoStatus.Information = Count;
		IoCompleteRequest(pIrp, IO_NO_INCREMENT);
	}
	KeReleaseSpinLockFromDpcLevel(&pLDI->Lock);
}
